"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const rxjs_1 = require("rxjs");
const operators_1 = require("rxjs/operators");
const message_1 = require("./message");
/**
 * Posts a message to a Window.
 * This method does not return anything, it's a send and forget API.
 * @param destinationWindow
 * @param message
 */
function postMessage(destinationWindow, message) {
    destinationWindow.postMessage(message, "*");
}
/**
 * Makes a Subject which can be used to send and receive post messages.
 * Note that this posts messages to a single `destinationWindow` but recieves
 * messages from any window.
 * @param thisWindow window of the current context.
 * @param destinationWindow window to post messages to.
 */
function MakePostMessageSubject(thisWindow, destinationWindow) {
    // Create an observe which will send the correct message given the Observable
    // event. e.g. an error will trigger an error message to be posted.
    const outboundMessages = {
        next: (payload) => postMessage(destinationWindow, {
            id: message_1.POST_MESSAGE_IDENTIFIER,
            type: message_1.PostMessageType.REGULAR_MESSAGE,
            payload
        }),
        error: (error) => postMessage(destinationWindow, {
            id: message_1.POST_MESSAGE_IDENTIFIER,
            type: message_1.PostMessageType.ERROR_MESSAGE,
            error
        }),
        complete: () => postMessage(destinationWindow, {
            id: message_1.POST_MESSAGE_IDENTIFIER,
            type: message_1.PostMessageType.COMPLETE_MESSAGE
        })
    };
    // Extract the payload of inbound messages and handle obervable complete and
    // error events.
    const inboundPostMessages = rxjs_1.fromEvent(thisWindow, "message");
    const closeMessages = inboundPostMessages.pipe(operators_1.filter(message => message.type === message_1.PostMessageType.COMPLETE_MESSAGE));
    const inboundMessages = inboundPostMessages.pipe(
    // Only accept widget post messages.
    operators_1.filter(message => message && message.id === message_1.POST_MESSAGE_IDENTIFIER), 
    // Throw an error locally if an error message is received.
    operators_1.tap(message => {
        if (message.type === message_1.PostMessageType.ERROR_MESSAGE) {
            throw new Error(message.error);
        }
    }), 
    // Complete when a completion message is received.
    operators_1.takeUntil(closeMessages.pipe(operators_1.first())), 
    // Extract the payload of regular messages.
    operators_1.map(message => message.payload));
    return rxjs_1.Subject.create(outboundMessages, inboundMessages);
}
exports.MakePostMessageSubject = MakePostMessageSubject;
